"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var debug_1 = __importDefault(require("debug"));
var log = (0, debug_1.default)("stub");
function stubProxy(env, functionViaPointer, traceStub) {
    if (traceStub === void 0) { traceStub = false; }
    return new Proxy(env, {
        get: function (target, key) {
            if (key in target) {
                return Reflect.get(target, key);
            }
            var f = functionViaPointer(key);
            if (f != null) {
                log("using function via pointer for ", key);
                return f;
            }
            // we ALWAYS log creating the stub.  traceStub determines if we print when using the stub.
            log("creating stub", key);
            if (traceStub) {
                log("creating stub", key);
                return function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    logStubUse(key, args, traceStub == "first");
                    return 0;
                };
            }
            else {
                // faster to not trace or even check, obviously.
                return function () { return 0; };
            }
        },
    });
}
exports.default = stubProxy;
var stubUsed = new Set([]);
function logStubUse(functionName, args, firstOnly) {
    if (firstOnly) {
        if (stubUsed.has(functionName))
            return;
        stubUsed.add(functionName);
    }
    console.warn("WARNING: using stub", functionName, args, firstOnly ? " -- only showing once" : "");
}
//# sourceMappingURL=stub.js.map